"""Classic Ranking System Module

The module provides implementation of well know ranking methods.
"""
from rstt.ranking.ranking import Ranking, get_disamb
from rstt.ranking.datamodel import KeyModel, GaussianModel
from rstt.ranking.rating import GlickoRating
from rstt.ranking.inferer import Glicko, Elo, PlayerLevel, PlayerWinPRC, EventStanding
from rstt.ranking.observer import GameByGame, BatchGame, KeyChecker


# ------------------------- #
# --- Consensus Ranking --- #
# ------------------------- #
class BTRanking(Ranking):
    def __init__(self, name: str = '', players=None):
        """Consensus Ranking For the Bradley-Terry Model

        Ranking based on the player's level() method.
        This also work for Time varying player, inherited class from :class:`rstt.player.playerTVS.PlayerTVS`,
        But it needs to be updated manually everytime player's level is updated.

        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        players : _type_, optional
            SPlayer to add to the ranking, by default None

        .. warning::
            BTRanking validity is limited to Bradley-Terry like models and is not suited for simulation using 'None-transitive' level.

            Behaver that for player with 'hyper volatile' level (like the :class:`rstt.player.gaussian.GaussianPlayer` which has a level changing at each update),
            the BTRanking uses the level of the player in question during the last update, it does not check for a 'theoretical' or 'statistical' mean. 
        """
        super().__init__(name=name,
                         datamodel=KeyModel(factory=lambda x: x.level()),
                         backend=PlayerLevel(),
                         handler=KeyChecker(),
                         players=players)

    def forward(self, *args, **kwargs):
        """:meta private:"""
        self.handler.handle_observations(
            infer=self.backend, datamodel=self.datamodel)


# ------------------------- #
# --- Empirical Ranking --- #
# ------------------------- #
class WinRate(Ranking):  # !!! need Player (with history) but ranking is not generic
    def __init__(self, name: str = '', default: float = -1.0, players=None):
        """Ranking based on Win rate


        Ranking that tracks the winrate of :class:`rstt.player.player.Player`.
        The winrate his tracked 'automatically'.
        The update function does not take any parameters, win rate is computed directly with the player's game history.

        WinRate uses:

            1. :class:`rstt.ranking.datamodel.KeyModel` as datamodel
            2. :class:`rstt.ranking.inferer.PlayerWinPr` as backend
            3. :class:`rstt.ranking.observer.KeyChecker` as handler


        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        default : float, optional
            A default rating value for when player have no game in their history, by default -1.0
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """
        backend = PlayerWinPRC(default=default)
        super().__init__(name=name,
                         datamodel=KeyModel(
                             factory=lambda x: backend.win_rate(x)),
                         backend=backend,
                         handler=KeyChecker(),
                         players=players)


class SuccessRanking(Ranking):
    def __init__(self, name: str = ',', buffer: int = 1, nb: int = 1, players=None, default=None):
        """Merit Based Ranking

        Usefull to implement Ranking system like the one in  `tennis <https://en.wikipedia.org/wiki/ATP_rankings>`_ for example.

        SuccessRanking uses:

            1. :class:`rstt.ranking.datamodel.KeyModel` as datamodel
            2. :class:`rstt.ranking.inferer.EventStanding` as backend
            3. :class:`rstt.ranking.observer.KeyChecker` as handler

        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        buffer : int
            Backend parameter. The number of event to consider for the rating, starting from the last.
        nb : int
            Backend parameter. The actual number of event in the buffer to use for the ratings computation.
        default : Optional[Dict[int, float]], optional
            Backend Parameter. Mapping placement in event to points for the rating, by default None
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """
        super().__init__(name=name,
                         datamodel=KeyModel(template=int),
                         backend=EventStanding(buffer=buffer,
                                               nb=nb, default=default),
                         handler=KeyChecker(),
                         players=players)

    def add_event(self, *args, **kwargs):
        """Wrapper of :func:`rstt.ranking.inferer.EventStanding.add_event`
        """
        self.backend.add_event(*args, **kwargs)

    def remove_event(self, *args, **kwargs):
        """Wrapper of :func:`rstt.ranking.inferer.EventStanding.remove_event`"""
        self.backend.remove_event(*args, **kwargs)

    def forward(self, event=None, points=None, *args, **kwargs):
        """:meta private:"""
        if event:
            self.backend.add_event(event, points)
        self.handler.handle_observations(infer=self.backend,
                                         datamodel=self.datamodel,
                                         *args, **kwargs)


# ------------------------- #
# ---- Common Ranking ----- #
# ------------------------- #
class BasicElo(Ranking):
    def __init__(self, name: str = '', default: float = 1500.0, k: float = 20.0, lc: float = 400.0, players=None):
        """Simple Elo System

        Impement a very simple ELo rating system.

        BasicElo uses:

            1. :class:`rstt.ranking.datamodel.KeyModel` as datamodel
            2. :class:`rstt.ranking.inferer.Elo` as backend
            3. :class:`rstt.ranking.observer.GameByGame` as handler

        .. note::
            In this implementation, the K value is a constant.
            In real world application, K can vary based on player's experience, or match's relevance.

        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        default : float, optional
            Datamodel parameter, a default elo rating, by default 1500.0
        k : float, optional
            Backend parameter, the K value, by default 20.0
        lc : float, optional
            Backend parameter, constant dividing the ratings difference in the expected score formula , by default 400.0
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """
        super().__init__(name=name,
                         datamodel=KeyModel(default=default),
                         backend=Elo(k=k, lc=lc),
                         handler=GameByGame(),
                         players=players)

    # def predict(self, game):
    #    _, teams_as_ratings, _, _ = self.handler.match_formating(datamodel=self.datamodel, game=game)
    #    r1 = teams_as_ratings[0][0]
    #    r2 = teams_as_ratings[1][0]
    #    return self.backend.expectedScore(rating1=r1, rating2=r2)


class BasicGlicko(Ranking):
    def __init__(self, name: str = '', handler=BatchGame(), mu: float = 1500.0, sigma: float = 350.0, players=None):
        """Simple Glicko system

        Implement A glicko rating system as originaly `proposed <https://www.glicko.net/glicko/glicko.pdf>`_.

        BasicGlicko uses:

            1. :class:`rstt.ranking.datamodel.GaussianModel` as datamodel
            2. :class:`rstt.ranking.inferer.Glicko` as backend
            3. :class:`rstt.ranking.observer.BatchGame` as handler

        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        handler : _type_, optional
            Backend as parameter, by default BatchGame()
            The original recommendation is to update the ranking by grouping matches within rating period.
            Which is what the BatchGame Observer do, (each update call represent one period). To match other glicko, use A GameByGame observer
        mu : float, optional
            Datamodel parameter, the default mu of the rating, by default 1500.0
        sigma : float, optional
           Datamodel parameter, the default sigma of the rating, by default 350.0
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """
        super().__init__(name=name,
                         datamodel=GaussianModel(
                             default=GlickoRating(mu, sigma)),
                         backend=Glicko(),
                         handler=handler,
                         players=players)

    @get_disamb
    def __step1(self):
        # TODO: check which player iterator to use
        for player in self:
            rating = self.datamodel.get(player)
            rating.sigma = self.backend.prePeriod_RD(rating)

    def forward(self, *args, **kwargs):
        self.__step1()
        self.handler.handle_observations(
            infer=self.backend, datamodel=self.datamodel, *args, **kwargs)

    # def predict(self, game):
    #    _, teams_as_ratings, _, _ = GameByGame().match_formating(datamodel=self.datamodel, game=game)
    #    r1 = teams_as_ratings[0][0]
    #    r2 = teams_as_ratings[1][0]
    #    return self.backend.expectedScore(rating1=r1, rating2=r2)


class BasicOS(Ranking):
    def __init__(self, name: str = '', model=None, players=None):
        """Simple OpenSkill Integretion

        Ranking to integrate an `openskill <https://openskill.me/en/stable/manual.html>`_ model into the rstt package.

        BasicOS uses:

            1. :class:`rstt.ranking.datamodel.GaussianModel` as datamodel to store model.rating instances.
            2. One of openskill models passed as parameter.
            3. :class:`rstt.ranking.observer.GameByGame` as handler


        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        model : openskills.models
            One of openskills.models implementation, by default None
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None


        Example:
        --------
        .. code-block:: python
            :linenos:

            from rstt import Player, BasicOS
            from openskill.models import PlackettLuce

            competitors = Player.create(nb=10)
            pl = BasicOS(name='Plackett-Luce', model= PlackettLuce(), players=competitors)
            pl.plot()
        """
        super().__init__(name=name,
                         datamodel=GaussianModel(
                             factory=lambda x: model.rating(name=x.name)),
                         backend=model,
                         handler=GameByGame(),
                         players=players)

    def quality(self, game) -> float:
        """Wrapper arround openskill model predict_draw function"""
        _, teams_as_ratings, _, _ = self.handler.match_formating(
            datamodel=self.datamodel, game=game)
        return self.backend.predict_draw(teams_as_ratings)

    # def predict(self, game) -> float:
    #    _, teams_as_ratings, _, _ = self.handler.match_formating(datamodel=self.datamodel, game=game)
    #    return self.backend.predict_win(teams_as_ratings)[0]



"""Module for Inferer.

Inferer are protocol providing a rate() method for ranking to update ratings.
"""

from typing import List, Dict, Tuple, Union, Optional, Any
from typeguard import typechecked

import rstt.config as cfg
from rstt.player import Player
from rstt.stypes import SPlayer, Event

import numpy as np
import math
import copy

import warnings


'''
    FIXME:
    - There is no consensus on the return type of the rate() functions
    - It is fine regarding the Protocol, but realy anoying when combine with Observer
'''


# ------------------------ #
# --- Player as Rating --- #
# ------------------------ #
class PlayerLevel:
    @typechecked
    def rate(self, player: SPlayer, *args, **kwars) -> Dict[SPlayer, float]:
        """Inferer based on players's level.

        Rates players on the returned value of the level() method. Usefull for 'Consensus' ranking variation.

        Parameters
        ----------
        player : SPlayer
            A player to rate

        Returns
        -------
        Dict[Player, float]
            the player and its new rating.
        """
        return {player: player.level()}


class PlayerWinPRC:
    def __init__(self, default: float = -1.0, scope: int = np.iinfo(np.int32).max):
        """Inferer based on Player win rate


        Parameters
        ----------
        default : float, optional
            A rating for when no game was yet played, by default -1.0
        scope : int, optional
            The number of game to consider, starting from the most recent one, by default np.iinfo(np.int32).max.
        """
        self.default = default
        self.scope = scope

    @typechecked
    def rate(self, player: Player, *args, **kwargs) -> Dict[Player, float]:
        """Win rate inference

        Parameters
        ----------
        player : Player
            a player to rate

        Returns
        -------
        Dict[Player, float]
            the player and its associated rating
        """
        return {player: self._win_rate(player)}

    def _win_rate(self, player: SPlayer):
        games = player.games()
        if games:
            if self.scope:
                games = games[-self.scope:]
            nb_wins = sum([1 for game in games if player is game.winner()])
            total = len(games)
            winrate = nb_wins / total * 100
        else:
            winrate = self.default
        return winrate


# ------------------------ #
# ----- Event Based ------ #
# ------------------------ #
class EventStanding:
    def __init__(self, buffer: int, nb: int, default: Optional[Dict[int, float]] = None):
        """Inferer tracking performances in tournaments

        Can be usefull to build ranking emulating one like in tennis or formula1.

        Parameters
        ----------
        buffer : int
            The number of event to consider for the rating, starting from the last.
        nb : int
            The actual number of event in the buffer to use for the ratings computation.
        default : Optional[Dict[int, float]], optional
            Mapping placement in event to points for the rating, by default None


        Example
        -------
            >>> EventStanding(buffer=10, nb=5, default={i:100-i for i in range(100)})

            Will rate each player based on his 5 best performances in the last 10 events added to the Inferer.
            Best performances in the sense of most points for his placements, not highest place (which could be different).

        """
        self.buffer = buffer
        self.nb = nb

        self.events = []
        self.points = {}

        self.__default_points = default if default is not None else cfg.EVENTSTANDING_DEFAULT_POINTS

    @typechecked
    def add_event(self, event: Union[Event, str], points: Optional[Dict[int, float]] = None):
        """Add an event to consider for ratings

        Event are always append as the last, thus will stay longer in the buffer than all previous eevnts added.

        Parameters
        ----------
        event : Union[Event, str]
            An event or its identifier.
        points : Optional[Dict[int, float]], optional
            A dictionnary mapping placement with points, by default None. In this case the default one is used for the added event.
        """
        points = points if points else self.__default_points
        if isinstance(event, str):  # by str
            self.events.append(event)
            self.points[event] = points
        else:  # by Event
            self.events.append(event.name())
            self.points[event.name()] = points

    @typechecked
    def remove_event(self, event: Union[Event, str]):
        """Remove an event

        Parameters
        ----------
        event : Union[Event, str]
            The event or its identifier, to remove from ratings considerations.
        """
        if isinstance(event, str):  # by str
            self.events.remove(event)
            del self.points[event]
        else:  # by Event
            self.events.remove(event.name())
            del self.points[event.name()]

    @typechecked
    def rate(self, player: Player) -> Dict[Player, float]:
        """Rate player based on performances.

        Parameters
        ----------
        player : Player
            A player to rate

        Returns
        -------
        Dict[Player, float]
            the player and its new rating.
        """
        # events that matter
        events = self.events[-self.buffer:]

        # collected points in events
        results = []
        for achievement in player.achievements():
            if achievement.event_name in events:
                results.append(
                    self.points[achievement.event_name][achievement.place])

        # get only the best results
        results.sort()
        best_results = results[-min(len(results), self.nb):]

        # sum the best results considered
        points = sum(best_results)
        return {player: points}


# ------------------------ #
# --- Game Score Based --- #
# ------------------------ #
class Elo:
    def __init__(self, k: float = 20.0, lc: float = 400.0):
        """Eo Inferer

        Simple implementation based on `wikipedia <https://en.wikipedia.org/wiki/Elo_rating_system#Theory>`_

        Parameters
        ----------
        k : float, optional
            The K-factor, by default 20.0
        lc : float, optional
            The constant dividing the ratings difference in the expected score formula, by default 400.0.
        """
        self.lc = lc
        self.K = k
        # TODO self.distribution = dist & change expectedScore

    @typechecked
    def rate(self, groups: List[List[float]], scores: List[float], *args, **kwars) -> List[List[float]]:
        """Rate method for elo

        Parameters
        ----------
        groups : List[List[float]]
            Elo ratings formated by teams, for example [[elo_player1], [elo_player2]].
        scores : List[float]
            corresponding scores of the ratings, for example [[1.0],[0.0]] assuming player1 won the duel.

        Returns
        -------
        List[List[float]]
            updated ratings in the formats [[new_elo1][new_elo2]]
        """
        # NOTE: groups: [[winner_elo][loser_elo]], scores [[1.0][0.0]]
        # FIXME: Current implementation does not seem to support a BatchGame observer
        # TODO: add errors for bad params

        # unpack args
        r1, r2 = groups[0][0], groups[1][0]
        s1, s2 = scores
        # cumpute new ratings
        new_r1 = self.update_rating(r1, r2, s1)
        new_r2 = self.update_rating(r2, r1, s2)
        return [[new_r1], [new_r2]]

    @typechecked
    def expectedScore(self, rating1: float, rating2: float) -> float:
        """Compute the expected score

        Parameters
        ----------
        rating1 : float
            a rating
        rating2 : float
            another rating

        Returns
        -------
        float
            expected result of the player with rating1 against the player with rating2
        """
        # TODO: 'DRY' -> use uf.logistic_elo
        return 1.0 / (1.0 + math.pow(10, (rating2-rating1)/self.lc))

    def update_rating(self, rating1: float, rating2: float, score: float):
        """Rating update

        Parameters
        ----------
        rating1 : float
            a rating
        rating2 : float
            another rating
        score : float
            the score associated to rating1

        Returns
        -------
        float
            the 'updated rating1'
        """
        expected_result = self.expectedScore(rating1, rating2)
        return rating1 + (self.K * (score-expected_result))


class Glicko:
    @typechecked
    def __init__(self, minRD: float = 30.0,
                 maxRD: float = 350.0,
                 c: float = 63.2,
                 q: float = math.log(10, math.e)/400,
                 lc: int = 400):
        """Glicko Inferer

        The `Glicko <https://en.wikipedia.org/wiki/Glicko_rating_system>`_ rating system is often described as an improvement of :class:`rstt.ranking.inferer.Elo`.
        here, the implementation is based on Dr. Mark E. Glickman `description <https://www.glicko.net/glicko/glicko.pdf>`_.

        .. note::
            The source paper gives more instruction (notion of rating period) than what an Inferer class should do in RSTT.
            Step1, for example is implemented by the :class:`rstt.ranking.standard.BasicGlicko`
            because it is related to the usage of the system, rather than what the Inferer does.

        .. warning::
            There is no type-checker support for 'Glicko ratings'.
            In the documentation we use the typehint 'GlickoRating'.
            Anything with a public mu and sigma attribute fits the bill.


        Parameters
        ----------
        minRD : float, optional
            minimal value of RD, by default 30.0
        maxRD : float, optional
            maximal value of RD, by default 350.0
        c : float, optional
            constant used for 'inactivity decay', by default 63.2
        q : float, optional
            No idea what it represent, feel free to play arround, by default math.log(10, math.e)/400
        lc : int, optional
            Logistic constant similar to the one in :class:`rstt.rnaking.inferer.Elo`, by default 400
        """

        # model constant
        self.__maxRD = maxRD
        self.__minRD = minRD
        self.lc = lc
        self.C = c
        self.Q = q

    def G(self, rd: float) -> float:
        """_summary_

        Implements: page 3, step2, g(RD) formula.

        Parameters
        ----------
        rd : float
            the RD of a rating

        Returns
        -------
        float
            g(RD)
        """
        return 1 / math.sqrt(1 + 3*self.Q*self.Q*(rd*rd)/(math.pi*math.pi))

    def expectedScore(self, rating1, rating2, update: bool = True) -> float:
        """Compute the expected score

        Implements: page 4, E(s|r,rj,RDj) when update=True
        or page 5, E otherwise.

        Parameters
        ----------
        rating1 : GlickoRating
            'main' rating
        rating2 : GlickoRating
            opponents rating
        update : bool, optional
            Wheter to use the formula for update or not, by default True.

        Returns
        -------
        float
            The expected score of the player with rating1 against player with rating2
        """
        RDi = 0 if update else rating1.sigma
        RDj = rating2.sigma
        ri, rj = rating1.mu, rating2.mu
        return 1 / (1 + math.pow(10, -self.G(math.sqrt(RDi*RDi + RDj*RDj)) * (ri-rj)/400))

    def d2(self, rating1, games: List[Tuple[Any, float]]) -> float:
        """
        Implements: page 4, d^2 formula.

        Parameters
        ----------
        rating1 : GlickoRating
            the main rating
        games : List[Tuple[GlickoRating, float]]
            A list of [opponent_rating, score_of_rating1]

        Returns
        -------
        float
            the d2 value

        Warns
        -----
            Rarely a ZeroDivisionError occurs. In this case, the warning contains all the computational information.
            Execution continues using a very small value instead.
        """
        all_EJ = []
        all_GJ = []
        for rating2, score in games:
            # get needed variables
            Ej = self.expectedScore(rating1, rating2, update=True)
            RDj = rating2.sigma
            Gj = self.G(RDj)

            # store vairables
            all_EJ.append(Ej)
            all_GJ.append(Gj)

        # big sum
        bigSum = 0
        for Gj, Ej, in zip(all_GJ, all_EJ):
            bigSum += Gj*Gj*Ej*(1-Ej)

        '''
        NOTE:
        Try/Expect is not part of the Glicko official algorithm  presentation.
        But I have encountered Unexpected ZeroDivisionError
        
        This is easly fixed by:
        return 1 / min( self.Q*self.Q*bigSum, lower_bound)
        
        However I could note find any specfic details about the choice of the boundary.
        
        Analytically, the term can not be equal to 0.0, it is always >0.
        Nnumercialy, it happens in extreme situation i.e does not arise in standard 'intended' Glicko usage.
        
        The package is for scientifical experimentation,
        It allows extreme case exploration and can not hide arbitrary choices.

        # !!! Do not fix unless it is possible to link a scientifical source justifying the implementation
        '''
        try:
            # d2 formula
            return 1 / (self.Q*self.Q*bigSum)
        except ZeroDivisionError:
            # !!! BUG: ZeroDivisionError observed with extreme rating differences
            # !!! this will now print variable of interest
            # !!! but code will run assuming maximal and mininal expected value possible between 0 and 1

            # HACK: just assume a very low 'bigSum'
            bigSum = 0.00000000001
            correction = 1 / (self.Q*self.Q*bigSum)

            msg = f"Glicko d2 ERROR: {rating1}, {games}\n {bigSum}, {all_EJ}, {all_GJ}\n d2 return value as been adjusted to 1/{bigSum}"
            warnings.warn(msg, RuntimeWarning)
            return correction

    # TODO: how to typecked
    def prePeriod_RD(self, rating: Any) -> float:
        """pre update RD value

        Implements: page 3, step1, formula (b).

        Parameters
        ----------
        rating : GlickoRating
            A rating to 'pre-update'

        Returns
        -------
        float
            the new RD value of the rating. 
        """
        new_RD = math.sqrt(rating.sigma*rating.sigma + self.C*self.C)
        # check boundaries on sigma - ??? move max() elsewhere
        return max(min(new_RD, self.__maxRD), self.__minRD)

    def newRating(self, rating1, games: List[Tuple[Any, float]]):
        """Rating Update method

        Implements: page 3, step2.

        Parameters
        ----------
        rating1 : GlickoRating
            a rating to update.
        games : List[Tuple[GlickoRating, float]]
            A list of results formated under as [opponent_rating, score_of rating1]

        Returns
        -------
        GlickoRating
            the new updated rating
        """

        # compute term 'a'
        d2 = self.d2(rating1, games)
        a = self.Q / ((1/(rating1.sigma*rating1.sigma)) + (1/d2))

        # lcompute term 'b'
        b = 0
        for rating2, score in games:
            b += self.G(rating2.sigma)*(score -
                                        self.expectedScore(rating1, rating2, update=True))

        # create new rating object to avoid 'side effect'
        rating = copy.copy(rating1)
        # post Period R
        rating.mu += a*b
        # post Period RD
        rating.sigma = math.sqrt(1/((1/rating1.sigma**2) + (1/d2)))

        return rating

    # FIXME: Does not support GameByGame observer
    def rate(self, rating, ratings: List[Any], scores: List[float], *args, **kwars):
        """Glicko rate method

        End to end method to compute a new glicko rating based on a collection of results

        Parameters
        ----------
        rating : GlickoRating
            the rating to update
        ratings : List[GlickoRating]
            list of opponent ratings
        scores : List[float]
            list of score achieved by rating1 against the 'ratings' opponents, in the same order

        Returns
        -------
        GlickoRating
            The new rating.
        """

        # formating
        games = [(r, s) for r, s in zip(ratings, scores)]
        return self.newRating(rating, games)



"""Module for Observer

Observer are protocol providing a handle_observations() method for ranking.
This method is responsible:

    - to proprocess observations - parameters passed to :func:`rstt.ranking.ranking.Raning.update` justifying ratings changes.
    - properly call the ranking.backend :func:`rstt.stypes.Infere.rate`
    - store new ratings in the ranking.datamodel :class:`rstt.stypes.RatingSystem`
    
.. warning::
    Reminder, currently RSTT support SMatch format is limited to the Duel class.
    This is also true for 'Game based' Observer even. Due to modularity, type checkers look for SMatch and not for Duel
    This may cause unexpected errors in your simulations.

"""

from typeguard import typechecked
from typing import List, Dict, Union, Tuple, Any, Optional

from rstt.stypes import SMatch, SPlayer, Inference, RatingSystem, Score
import rstt.utils.utils as uu

# QUEST: Do Observer realy need to be typechecked ?


@typechecked
def assign_ratings(datamodel: RatingSystem, ratings: Dict[SPlayer, Any]):
    """Push updated ratings in the RatingSystem

    Utility function to 'push' updated ratings into the raking's datamodel.

    Parameters
    ----------
    datamodel : RatingSystem
        The datamodel of a ranking.
    ratings : Dict[SPlayer, Any]
        Updated ratings to store, usually returned values of a Inferer.rate method.

    :meta private:
    """
    for key, rating in ratings.items():
        datamodel.set(key, rating)


@typechecked
def match_data(game: SMatch, datamodel: RatingSystem) -> Tuple[List[List[SPlayer]], List[List[Any]], Score, List[int]]:
    """Extract Match infos

    Utility function to format match into suitable data.
    Return is suitable for any RSTT Game Score Based Inferer, but also for some external systems.

    Parameters
    ----------
    game : SMatch
        A match to extract information from
    datamodel : RatingSystem
        A datamodel to extract players rating's.

    Returns
    -------
    Tuple[List[List[SPlayer]], List[List[Any]], Score, List[int]]
        Collection of data as foloow:
        1. Participants: List[List[SPlayer]]
        2. Partitipants ratings in the same order: List[List[Any]]
        3. Team scores: Score
        4. Teams ranks. List[int]

    :meta private:
    """
    teams_as_players = game.teams()
    teams_as_ratings = [
        [datamodel.get(player) for player in team] for team in teams_as_players]

    if game.live():
        scores, ranks = None, None
    else:
        scores = game.scores()
        ranks = game.ranks()

    return teams_as_players, teams_as_ratings, scores, ranks


@typechecked
def players_ratings_to_dict(players: List[List[SPlayer]], ratings: List[List[Any]]) -> Dict[SPlayer, Any]:
    return {p: r for p, r in zip(uu.flatten(players), uu.flatten(ratings))}


class GameByGame:
    """Match based Observer"""
    @typechecked
    def handle_observations(self, infer: Inference,
                            datamodel: RatingSystem,
                            games: Optional[Union[SMatch,
                                                  List[SMatch]]] = None,
                            # FIXME: proper typing with Scheduler class WIP
                            event: Optional[Any] = None, *args, **kwargs):
        """Game by Game updating Procedure

        Implementing an iterative approach where each observations triggers the entire updating workflows.
        In particular, new ratings are stored inbetween of each iterations, and the prior ones are lost.

        Parameters
        ----------
        infer : Inference
            A system to compute new rating
        datamodel : RatingSystem
            A container of ratings
        games : Optional[Union[SMatch, List[SMatch]]], optional
            Observations justifying a ranking update, by default None
        event : Optional[Any], optional
            Alternative for the games parameter, must provide a games() method, by default None

        Raises
        ------
        ValueError
            Indicate an incompatible arg call, either provide 'event' or 'games' not both.
        """
        if event:
            if games:
                msg = f'Incompatible parameters, Only one of games or event must be passed'
                raise ValueError(msg)
            else:
                self.handle_observations(
                    infer=infer, datamodel=datamodel, games=event.games())
        elif isinstance(games, SMatch):
            self.single_game(infer, datamodel, games, *args, **kwargs)
        else:
            for game in games:
                self.single_game(infer, datamodel, game, *args, **kwargs)

    @typechecked
    def single_game(self, infer: Inference, datamodel: RatingSystem, game: SMatch, *args, **kwargs):
        """Implement to update workflow for a single match

        Parameters
        ----------
        infer : Inference
            A system to compute new rating
        datamodel : RatingSystem
            A container of ratings
        games : Optional[Union[SMatch, List[SMatch]]], optional
            An Observation justifying a ranking update, by default None

        :meta private:
        """
        players, ratings, score, ranks = match_data(game, datamodel)
        try:
            output = infer.rate(ratings, ranks=ranks, *args, **kwargs)
        except:
            output = infer.rate(ratings, score, *args, **kwargs)
        new_ratings = players_ratings_to_dict(players, output)
        assign_ratings(datamodel, new_ratings)


class KeyChecker:
    """Player based Observer"""

    def handle_observations(self, infer: Inference, datamodel: RatingSystem, *args, **kwargs):
        """No observation Observer

        Helps implementing Ranking in cases where an update can be justify even in the absence of tracebable triggers.
        For example, A consensus ranking for player with an evolving level.

        Additionnaly can help build Ranking based on player history:
            - Win rate based.
            - Career earnings leaderboard
            - ...

        Parameters
        ----------
        infer : Inference
            A system to compute new rating
        datamodel : RatingSystem
            A container of ratings
        """
        new_ratings = {}

        for player in datamodel.keys():
            new_ratings.update(infer.rate(player, *args, **kwargs))

        assign_ratings(datamodel, new_ratings)


class BatchGame:
    """Match Based Observer"""
    @typechecked
    def handle_observations(self, infer: Inference,
                            datamodel: RatingSystem,
                            games: Optional[Union[SMatch,
                                                  List[SMatch]]] = None,
                            # FIXME: proper typing with Scheduler class WIP
                            event: Optional[Any] = None, *args, **kwargs):
        """All Matches at once updating procedure

        Alternative to the :class:`rstt.ranking.observer.GamebyGame` observer. Some rating system, like Elo and Glicko
        support updates where all matches are considered at once for the rating update.

        In this workflows, ratings are stored after all matches have been processed. Every computation is performed using the prior ratings
        (i.e the one stored in the datamodel before the method call)

        Parameters
        ----------
        infer : Inference
            A system to compute new rating
        datamodel : RatingSystem
            A container of ratings
        games : Optional[Union[SMatch, List[SMatch]]], optional
            Observations justifying a ranking update, by default None
        event : Optional[Any], optional
            Alternative for the games parameter, must provide a games() method, by default None

        Raises
        ------
        ValueError
            Indicate an incompatible arg call, either provide 'event' or 'games' not both.
        """
        if event:
            if games:
                msg = f'Incompatible parameters, Only one of games or event must be passed'
                raise ValueError(msg)
            else:
                self.handle_observations(
                    infer=infer, datamodel=datamodel, games=event.games())
        else:
            games = self.to_list(games)
            players = self.involved_keys(games)
            new_ratings = {}
            for player in players:
                player_games = self.game_of_interest(player, games)
                ratings = [datamodel.get(game.opponent(player))
                           for game in player_games]
                scores = [self.player_score(player, game)
                          for game in player_games]
                rating = datamodel.get(player)
                new_ratings[player] = infer.rate(rating, ratings, scores)

            assign_ratings(datamodel, new_ratings)

    @typechecked
    def player_score(self, player: SPlayer, game: SMatch):
        """Exctact player score in a gievn macth

        :meta private:
        """
        if game.winner() == player:
            return 1.0
        elif game.loser() == player:
            return 0.0
        elif game.isdraw():
            return 0.5
        else:
            msg = f'Player {player} can not be asigned a score value (1.0, 0.5, 0.0) for Game {game}'
            raise ValueError(msg)

    @typechecked
    def game_of_interest(self, player: SPlayer, games: List[SMatch]) -> List[SMatch]:
        """Find all game where a given player participated in the batch

        :meta private:
        """
        return [game for game in games if player in game]

    @typechecked
    def involved_keys(self, games: List[SMatch]) -> List[SPlayer]:
        """Find all opponents of a gien player in the batch

        :meta private:
        """
        winners = [game.winner() for game in games]
        losers = [game.loser() for game in games]
        return list(set(winners+losers))

    @typechecked
    def to_list(self, games: Union[SMatch, List[SMatch]]) -> List[SMatch]:
        """input formater for the inferer.rate

        :meta private:
        """
        if not isinstance(games, list):
            games = [games]
        return games
